/* X-Chat
 * Copyright (C) 1998 Peter Zelezny.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * =========================================================================
 *
 * xtext, the text widget used by X-Chat.
 *
 * By Peter Zelezny <zed@linux.com>.
 * Some functions used from Zvt and Eterm (transparency stuff).
 *
 * TODO:
 *     - text marking (works line-by-line only)
 *     - word clicking
 *     - URL highlight (as you move the pointer)
 *     - Wide/16bit font support (someone else must do this!)
 *     - Exposure events redraw the whole page, optimize this.
 *     - word wrap
 *     - when a subline starts a page, parse previous attribs.
 *
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkselection.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include "xtext.h"
#ifdef USE_IMLIB
#include <gdk_imlib.h>
#endif

#define REFRESH_TIMEOUT 15

#undef GTK_WIDGET
#define GTK_WIDGET(n) ((GtkWidget*)n)

static GtkWidgetClass *parent_class = NULL;

static void gtk_xtext_render_page (GtkXText *xtext, int startline);
static void gtk_xtext_calc_lines (GtkXText *xtext);
static void gtk_xtext_load_trans (GtkXText *xtext);
static void gtk_xtext_free_trans (GtkXText *xtext);
static textentry *gtk_xtext_nth (GtkXText *xtext, textentry *start_ent, int line, int width, int *subline);
static gint gtk_xtext_selection_kill (GtkWidget *widget, GdkEventSelection *event);
static void gtk_xtext_selection_get (GtkWidget        *widget, 
			GtkSelectionData *selection_data_ptr,
			guint             info,
			guint             time);
static int gtk_xtext_text_width (GtkXText *xtext, unsigned char *text, int len);
static void gtk_xtext_adjustment_changed (GtkAdjustment *adj, GtkXText *xtext);
static char *gtk_xtext_strip_color (unsigned char *text, int len);


static void
gtk_xtext_init (GtkXText *xtext)
{
   xtext->old_value = -1;
   xtext->pixmap = NULL;
   xtext->text_first = NULL;
   xtext->text_last = NULL;
   xtext->io_tag = -1;
   xtext->frozen = 0;
   xtext->num_lines = 0;
   xtext->max_lines = 0;
   xtext->col_back = 0;
   xtext->col_fore = 1;
   xtext->nc = 0;
   xtext->bold = FALSE;
   xtext->underline = FALSE;
   xtext->reverse = FALSE;
   xtext->font = gdk_font_load ("fixed");

   xtext->adj = (GtkAdjustment *) gtk_adjustment_new (0, 0, 0, 1, 0, 0);
   gtk_signal_connect (GTK_OBJECT (xtext->adj), "value_changed",
                       GTK_SIGNAL_FUNC (gtk_xtext_adjustment_changed), xtext);
   gtk_signal_connect (GTK_OBJECT (xtext), "selection_clear_event",
                       GTK_SIGNAL_FUNC (gtk_xtext_selection_kill), xtext);
   gtk_selection_add_target (GTK_WIDGET (xtext),
                             GDK_SELECTION_PRIMARY,
                             GDK_SELECTION_TYPE_STRING,
                             1);
   gtk_signal_connect (GTK_OBJECT (xtext), "selection_get",
                       GTK_SIGNAL_FUNC (gtk_xtext_selection_get), xtext);
}

static void
gtk_xtext_adjustment_set (GtkXText *xtext)
{
   GtkAdjustment *adj = xtext->adj;

   adj->lower = 0;
   adj->upper = xtext->num_lines;

   adj->page_size = (GTK_WIDGET (xtext)->allocation.height - xtext->font->descent) / xtext->fontsize;
   adj->page_increment = adj->page_size;

   if (adj->value > (adj->upper - adj->page_size))
   {
      adj->value = (adj->upper - adj->page_size);
      gtk_adjustment_value_changed (adj);
   }

   gtk_adjustment_changed (adj);
}

static gint
gtk_xtext_adjustment_timeout (GtkXText *xtext)
{
   gtk_xtext_render_page (xtext, xtext->adj->value);
   xtext->io_tag = -1;
   return 0;
}

static void
gtk_xtext_adjustment_changed (GtkAdjustment *adj, GtkXText *xtext)
{
   if (xtext->frozen)
      return;

   if ((int)xtext->old_value != (int)xtext->adj->value)
   {

      if (xtext->adj->value+1 == xtext->old_value ||
          xtext->adj->value-1 == xtext->old_value)
      { 
         if (xtext->io_tag != -1)
         {
            gtk_timeout_remove (xtext->io_tag);
            xtext->io_tag = -1;
         }
         gtk_xtext_render_page (xtext, xtext->adj->value);
      } else
      {
         if (xtext->io_tag != -1)
            return;
         xtext->io_tag = gtk_timeout_add (REFRESH_TIMEOUT,
                                       (GtkFunction)gtk_xtext_adjustment_timeout,
                                       xtext);
      }
   }
   xtext->old_value = adj->value;
}

GtkWidget*
gtk_xtext_new (int indent, int separator)
{
   GtkXText *xtext;

   xtext = gtk_type_new (gtk_xtext_get_type ());
   xtext->indent = indent;
   xtext->separator = separator;

   return GTK_WIDGET (xtext);
}

static void
gtk_xtext_destroy (GtkObject *object)
{
   GtkXText *xtext;
   textentry *ent, *next;

//   g_return_if_fail (object != NULL);
//   g_return_if_fail (GTK_IS_XTEXT (object));

   xtext = GTK_XTEXT (object);

   if (xtext->io_tag != -1)
   {
      gtk_timeout_remove (xtext->io_tag);
   }

   if (xtext->pixmap)
   {
      if (xtext->transparent)
         gtk_xtext_free_trans (xtext);
      else
         gdk_pixmap_unref (xtext->pixmap);
   }

   if (xtext->font)
      gdk_font_unref (xtext->font);

   if (xtext->adj)
      gtk_object_destroy (GTK_OBJECT (xtext->adj));

   if (xtext->bgc)
      gdk_gc_destroy (xtext->bgc);

   if (xtext->fgc)
      gdk_gc_destroy (xtext->fgc);

   ent = xtext->text_first;
   while (ent)
   {
      next = ent->next;
      free (ent);
      ent = next;
   }

   if (GTK_OBJECT_CLASS (parent_class)->destroy)
      (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
gtk_xtext_realize (GtkWidget *widget)
{
   GtkXText *xtext;
   GdkWindowAttr attributes;

//   g_return_if_fail (widget != NULL);
//   g_return_if_fail (GTK_IS_XTEXT (widget));

   GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
   xtext = GTK_XTEXT (widget);

   attributes.x = widget->allocation.x;
   attributes.y = widget->allocation.y;
   attributes.width = widget->allocation.width;
   attributes.height = widget->allocation.height;
   attributes.wclass = GDK_INPUT_OUTPUT;
   attributes.window_type = GDK_WINDOW_CHILD;
   attributes.event_mask = gtk_widget_get_events (widget) | 
          GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK | 
          GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK;// |
      //    GDK_POINTER_MOTION_HINT_MASK;
   attributes.visual = gtk_widget_get_visual (widget);
   attributes.colormap = gtk_widget_get_colormap (widget);

   widget->window = gdk_window_new (widget->parent->window, &attributes,
                       GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP);

//   widget->style = gtk_style_attach (widget->style, widget->window);

   gdk_window_set_user_data (widget->window, widget);

//   gtk_style_set_background (widget->style, widget->window, GTK_STATE_ACTIVE);
   xtext->bgc = gdk_gc_new (widget->window);
   xtext->fgc = gdk_gc_new (widget->window);
//   gdk_window_set_background (widget->window, &widget->style->black);

   XSetFont (((GdkWindowPrivate *)GTK_WIDGET (xtext)->window)->xdisplay,
             ((GdkGCPrivate *)xtext->fgc)->xgc,
             xtext->xfont->fid);

   XSetForeground (((GdkGCPrivate*)xtext->fgc)->xdisplay,
                   ((GdkGCPrivate*)xtext->fgc)->xgc,
                   xtext->palette[0]);
   XSetBackground (((GdkGCPrivate*)xtext->fgc)->xdisplay,
                   ((GdkGCPrivate*)xtext->fgc)->xgc,
                   xtext->palette[1]);
   XSetForeground (((GdkGCPrivate*)xtext->bgc)->xdisplay,
                   ((GdkGCPrivate*)xtext->bgc)->xgc,
                   xtext->palette[1]);
   XSetWindowBackground (((GdkWindowPrivate*)GTK_WIDGET(xtext)->window)->xdisplay,
                         ((GdkWindowPrivate*)GTK_WIDGET(xtext)->window)->xwindow,
                         xtext->palette[1]);

   if (xtext->transparent)
      gtk_xtext_load_trans (xtext);
}

static void 
gtk_xtext_size_request (GtkWidget      *widget,
                        GtkRequisition *requisition)
{
   requisition->width = 500;
   requisition->height = 316;
}

static void
gtk_xtext_size_allocate (GtkWidget     *widget,
                         GtkAllocation *allocation)
{
   widget->allocation = *allocation;
   if (GTK_WIDGET_REALIZED (widget))
   {
      gdk_window_move_resize (widget->window,
                              allocation->x, allocation->y,
                              allocation->width, allocation->height);
   }
   gtk_xtext_calc_lines (GTK_XTEXT (widget));
}

int
gtk_xtext_moved (GtkXText *xtext)
{
   if (xtext->transparent)
   {
      gtk_xtext_free_trans (xtext);
      gtk_xtext_load_trans (xtext);
   }
   xtext->io_tag = -1;
   return 0;
}

static gint
gtk_xtext_expose (GtkWidget      *widget,
                  GdkEventExpose *event)
{
   gtk_xtext_render_page (GTK_XTEXT (widget), GTK_XTEXT (widget)->adj->value);

   if (GTK_XTEXT (widget)->io_tag == -1)
      GTK_XTEXT (widget)->io_tag = gtk_timeout_add (REFRESH_TIMEOUT,
                                       (GtkFunction)gtk_xtext_moved,
                                       widget);
   return FALSE;
}

static void
gtk_xtext_selection_clear (GtkXText *xtext)
{
   textentry *ent;

   ent = xtext->text_first;
   while (ent)
   {
      ent->mark_start = -1;
      ent->mark_end = -1;
      ent = ent->next;
   }
}

static int
gtk_xtext_selection_line (GtkXText *xtext, textentry *ent, char *str, int line, int lines_max, int subline, int indent, int str_width)
{
   int y;
   int width;
   int height;
   int orig_len;
   int len;
   int ret = 1;

   len = orig_len = strlen (str);

   gdk_window_get_size (GTK_WIDGET (xtext)->window, &width, &height);
   width--;

   y = (xtext->fontsize * line) + xtext->font->ascent;

   if (str_width == -1)
      str_width = gtk_xtext_text_width (xtext, str, len);

   str_width += indent;

   while (str_width > width)
   {
      len--;
      if (len == 0)
         return 1;
      str_width = gtk_xtext_text_width (xtext, str, len) + indent;
   }

   if (!subline)
   {
int i;
//      gtk_xtext_render_str (xtext, y, ent, str, len, width, indent);
for (i=0; i<len; i++)
printf("%c", str[i]);
printf("\n");

      if (ent->mark_start == -1)
         ent->mark_start = str - ent->str;
      ent->mark_end = (str - ent->str) + len;

   } else
   {
      subline--;
      line--;
      ret--;
   }

   if (len != orig_len && lines_max > line+1)
   {
      ret += gtk_xtext_selection_line (xtext, ent, str + len, line+1, lines_max, subline, xtext->indent, -1);
   }

   return ret;
}



static void
gtk_xtext_selection_draw (GtkXText *xtext, int moved)
{
   textentry *ent;

   int select_start_x;
   int select_start_y;
   int select_width;
   int select_height;

   int win_width, win_height;

   int line_start;
   int line_end;
   int lines_max;
   int subline;
   int offset;

   gdk_window_get_size (GTK_WIDGET (xtext)->window, &win_width, &win_height);
   win_width--;

   select_start_x = MIN(xtext->select_start_x, xtext->select_end_x);
   select_start_y = MIN(xtext->select_start_y, xtext->select_end_y);

   offset = xtext->adj->value - xtext->select_start_adj;

   select_width = abs(xtext->select_end_x - xtext->select_start_x);
   select_height = abs(xtext->select_end_y - xtext->select_start_y);

   lines_max = ((win_height - xtext->font->descent) / xtext->fontsize);

   line_start = ((select_start_y - xtext->font->descent) / xtext->fontsize);
   line_end = ((select_start_y + select_height - xtext->font->descent) / xtext->fontsize) + 1;

   if (line_end > lines_max)
      line_end = lines_max;

   gtk_xtext_selection_clear (xtext);

   ent = gtk_xtext_nth (xtext, xtext->text_first,
                         line_start + xtext->adj->value,
                         win_width, &subline);

printf("\033[1H\033[2J\033[1H--------\n");
   while (ent)
   {
      line_start += gtk_xtext_selection_line (xtext, ent, ent->str, line_start, line_end, subline, 0, ent->str_width); 
      subline = 0;

      if (line_start >= line_end)
         break;

      ent = ent->next;
   }
printf("----end-----\n");

   gtk_xtext_render_page (xtext, xtext->adj->value);
}

static void
gtk_xtext_selection_update (GtkXText *xtext, GdkEventMotion *event)
{
   int y, height;
   int win_width, win_height;
   int moved;

   gdk_window_get_size (GTK_WIDGET (xtext)->window, &win_width, &win_height);
   win_width--;

   moved = xtext->adj->value - xtext->select_start_adj;  /* moved since start */

   gtk_xtext_selection_draw (xtext, moved);

   y = MIN(xtext->select_start_y, xtext->select_end_y);

   height = xtext->select_end_y - xtext->select_start_y;
   if (height < 0)
      height *= -1;

   /* selecting past top of window, scroll up! */
   if (y < 0 && xtext->adj->value >= 0)
   {
      xtext->adj->value--;
      gtk_adjustment_changed (xtext->adj);
   }

   /* selecting past bottom of window, scroll down! */
   if (y+height > win_height &&
       xtext->adj->value < (xtext->adj->upper - xtext->adj->page_size))
   {
      xtext->adj->value++;
      gtk_adjustment_changed (xtext->adj);
   }

   gtk_selection_owner_set (GTK_WIDGET (xtext), GDK_SELECTION_PRIMARY, event->time);
}

static gint
gtk_xtext_motion_notify (GtkWidget      *widget,
			                GdkEventMotion *event)
{
   GdkModifierType mask;
   GtkXText *xtext = GTK_XTEXT (widget);

   if (xtext->button_down)
   {
      gtk_grab_add (widget);
      gdk_pointer_grab (widget->window, TRUE,
			GDK_BUTTON_RELEASE_MASK |
			GDK_BUTTON_MOTION_MASK,// |
//         GDK_POINTER_MOTION_HINT_MASK,
		   NULL, NULL, 0);
      gdk_window_get_pointer (widget->window, &xtext->select_end_x,
                              &xtext->select_end_y, &mask);
      gtk_xtext_selection_update (xtext, event);
   }

   return FALSE;
}

static gint
gtk_xtext_button_release (GtkWidget      *widget,
			                 GdkEventButton *event)
{
   GtkXText *xtext = GTK_XTEXT (widget);

   if (event->button == 1)
   {
      xtext->button_down = FALSE;

      gtk_grab_remove (widget);
      gdk_pointer_ungrab (0);

      if (xtext->select_start_x == event->x &&
          xtext->select_start_y == event->y)
      {
         gtk_xtext_selection_clear (xtext);
         gtk_xtext_render_page (xtext, xtext->adj->value);
      }
   }/* else
   {
      gtk_xtext_selection_clear (xtext);
      gtk_xtext_render_page (xtext, xtext->adj->value);
   }*/

   return FALSE;
}

static gint
gtk_xtext_button_press (GtkWidget      *widget,
			               GdkEventButton *event)
{
   GtkXText *xtext = GTK_XTEXT (widget);
   GdkModifierType mask;

   if (event->button != 1)    /* we only want left button */
      return FALSE;

   if (event->type == GDK_2BUTTON_PRESS)
   {
      puts ("UNIMP: word select");
      return FALSE;
   }

   if (event->type == GDK_3BUTTON_PRESS)
   {
      puts ("UNIMP: line select");
      return FALSE;
   }

   xtext->button_down = TRUE;

   gdk_window_get_pointer (widget->window, &xtext->select_start_x,
                           &xtext->select_start_y, &mask);

   xtext->select_start_adj = xtext->adj->value;

   return FALSE;
}

/* another program has claimed the selection */

static gint
gtk_xtext_selection_kill (GtkWidget *widget, GdkEventSelection *event)
{
   gtk_xtext_selection_clear (GTK_XTEXT (widget));
   gtk_xtext_render_page (GTK_XTEXT (widget), GTK_XTEXT (widget)->adj->value);
   return TRUE;
}

/* another program is asking for our selection */

static void
gtk_xtext_selection_get (GtkWidget        *widget, 
			GtkSelectionData *selection_data_ptr,
			guint             info,
			guint             time)
{
   GtkXText *xtext = GTK_XTEXT (widget);
   textentry *ent;
   char *txt;
   char *pos;
   char *stripped;
   char *st;
   int len;

   len = 0;
   ent = xtext->text_first;
   while (ent)
   {
      if (ent->mark_start != -1)
      {
         len += ent->mark_end - ent->mark_start;
      }
      ent = ent->next;
   }

   pos = txt = malloc (len);
   ent = xtext->text_first;
   while (ent)
   {
      if (ent->mark_start != -1)
      {
         memcpy (pos, ent->str + ent->mark_start, ent->mark_end - ent->mark_start);
         pos += ent->mark_end - ent->mark_start;
      }
      ent = ent->next;
   }

   st = stripped = gtk_xtext_strip_color (txt, len);
   while (*stripped == ' ')
      stripped++;

   gtk_selection_data_set (selection_data_ptr, GDK_SELECTION_TYPE_STRING,
			                  8, stripped, strlen (stripped));

   free (st);
   free (txt);
}

static void
gtk_xtext_class_init (GtkXTextClass *class)
{
   GtkObjectClass *object_class;
   GtkWidgetClass *widget_class;

   object_class = (GtkObjectClass*) class;
   widget_class = (GtkWidgetClass*) class;

   parent_class = gtk_type_class (gtk_widget_get_type ());

   object_class->destroy = gtk_xtext_destroy;

   widget_class->realize = gtk_xtext_realize;
   widget_class->expose_event = gtk_xtext_expose;
   widget_class->size_request = gtk_xtext_size_request;
   widget_class->size_allocate = gtk_xtext_size_allocate;
   widget_class->button_press_event = gtk_xtext_button_press;
   widget_class->button_release_event = gtk_xtext_button_release;
   widget_class->motion_notify_event = gtk_xtext_motion_notify;
}

guint
gtk_xtext_get_type ()
{
   static guint xtext_type = 0;

   if (!xtext_type)
   {
      GtkTypeInfo xtext_info =
      {
         "GtkXText",
         sizeof (GtkXText),
         sizeof (GtkXTextClass),
         (GtkClassInitFunc) gtk_xtext_class_init,
         (GtkObjectInitFunc) gtk_xtext_init,
         (GtkArgSetFunc) NULL,
         (GtkArgGetFunc) NULL,
      };

      xtext_type = gtk_type_unique (gtk_widget_get_type (), &xtext_info);
   }

   return xtext_type;
}

void
gtk_xtext_thaw (GtkXText *xtext)
{
   if (xtext->frozen > 0)
      xtext->frozen--;

   if (xtext->frozen == 0)
      gtk_xtext_render_page (xtext, xtext->adj->value);
}

void
gtk_xtext_freeze (GtkXText *xtext)
{
   xtext->frozen++;
}

/* strip MIRC colors and other attribs. */

static char *
gtk_xtext_strip_color (unsigned char *text, int len)
{
   int nc = 0;
   int i = 0;
   int col = FALSE;
   char *new_str = malloc (len + 2);

   while (len > 0)
   {
      if ((col && isdigit (*text) && nc < 2) ||
          (col && *text == ',' && nc < 3))
      {
         nc++;
         if (*text == ',')
            nc = 0;
      } else
      {
         if (col)
            col = FALSE;
         switch (*text)
         {
         case ATTR_COLOR:
            col = TRUE;
            nc = 0;
            break;
         case ATTR_BEEP:
         case ATTR_RESET:
         case ATTR_REVERSE:
         case ATTR_BOLD:
         case ATTR_UNDERLINE:
            break;
         default:
            new_str[i] = *text;
            i++;
         }
      }
      text++;
      len--;
   }

   new_str[i] = 0;

   return new_str;
}

/* gives width of a string - with no mIRC codes in it */

static int
gtk_xtext_text_width_simple (GtkXText *xtext, unsigned char *str, int len)
{
   int width = 0;

   while (len)
   {
      width += xtext->fontwidth[*str];
      len--;
      str++;
   }

   return width;
}

/* gives width of a string, excluding the mIRC codes */

static int
gtk_xtext_text_width (GtkXText *xtext, unsigned char *text, int len)
{
   int nc = 0;
   int width = 0;
   int col = FALSE;

   while (len > 0)
   {

      if ((col && isdigit (*text) && nc < 2) ||
          (col && *text == ',' && nc < 3))
      {
         nc++;
         if (*text == ',')
            nc = 0;
      } else
      {
         if (col)
            col = FALSE;
         switch (*text)
         {
         case ATTR_COLOR:
            col = TRUE;
            nc = 0;
            break;
         case ATTR_BEEP:
         case ATTR_RESET:
         case ATTR_REVERSE:
         case ATTR_BOLD:
         case ATTR_UNDERLINE:
            break;
         default:
            if (xtext->fonttype == FONT_1BYTE)
               width += xtext->fontwidth[*text];
            else
            {
               width += gdk_char_width_wc (xtext->font, *((GdkWChar*)text));
               //width += gdk_text_width (xtext->font, text, 1);
               len--;
               text++;
            }
         }
      }
      text++;
      len--;
   }

   return width;
}

/* actually draw text to screen */

static int
gtk_xtext_render_flush (GtkXText *xtext, int x, int y, char *str, int len, GdkGC *col)
{
   GdkWindowPrivate *drawable_private;
   GdkGCPrivate *gc_private;
   int str_width;
   int dofill;

   if (xtext->fonttype == FONT_1BYTE)
      str_width = gtk_xtext_text_width_simple (xtext, str, len);
   else
      str_width = gdk_text_width (xtext->font, str, len);

   drawable_private = (GdkWindowPrivate *)GTK_WIDGET(xtext)->window;
   gc_private = (GdkGCPrivate *)col;

   if (!xtext->backcolor && xtext->pixmap)
   {
      dofill = FALSE;
      /*gdk_draw_rectangle (GTK_WIDGET (xtext)->window,
                          xtext->bgc,
                          1,
                          x, y - (xtext->font->ascent),
                          str_width - x, xtext->fontsize);*/
      //if (xtext->transparent)
         XFillRectangle (drawable_private->xdisplay,
                         drawable_private->xwindow,
                         ((GdkGCPrivate *)xtext->bgc)->xgc,
                         x, y - (xtext->font->ascent),
                         str_width - x, xtext->fontsize);
      /*else
         XClearArea (drawable_private->xdisplay,
                     drawable_private->xwindow,
                     x, y - (xtext->font->ascent),
                     str_width - x, xtext->fontsize, False);*/
   } else
      dofill = TRUE;

   switch (xtext->fonttype)
   {
      case FONT_1BYTE:
         if (dofill)
            XDrawImageString (drawable_private->xdisplay,
                              drawable_private->xwindow,
		                        gc_private->xgc, x, y, str, len);
         else
            XDrawString (drawable_private->xdisplay,
                         drawable_private->xwindow,
		                   gc_private->xgc, x, y, str, len);
         if (xtext->bold)
            XDrawString (drawable_private->xdisplay, drawable_private->xwindow,
		                   gc_private->xgc, x+1, y, str, len);
         break;

      case FONT_2BYTE:
         len /= 2;
         if (dofill)
            XDrawImageString16 (drawable_private->xdisplay,
                                drawable_private->xwindow,
		                          gc_private->xgc, x, y, (XChar2b *)str, len);
         else
            XDrawString16 (drawable_private->xdisplay,
                           drawable_private->xwindow,
		                     gc_private->xgc, x, y, (XChar2b *)str, len);
         if (xtext->bold)
            XDrawString16 (drawable_private->xdisplay,
                           drawable_private->xwindow,
		                     gc_private->xgc, x+1, y, (XChar2b *)str, len);
         break;

      case FONT_SET:
         if (dofill)
            XmbDrawImageString (drawable_private->xdisplay,
                                drawable_private->xwindow,
                                (XFontSet)xtext->xfont, gc_private->xgc,
                                x, y, str, len);
         else
            XmbDrawString (drawable_private->xdisplay,
                           drawable_private->xwindow,
                           (XFontSet)xtext->xfont, gc_private->xgc,
                           x, y, str, len);
         if (xtext->bold)
            XmbDrawString (drawable_private->xdisplay,
                           drawable_private->xwindow,
                           (XFontSet)xtext->xfont, gc_private->xgc,
                           x+1, y, str, len);
   }

   if (xtext->underline)
      XDrawLine (drawable_private->xdisplay, drawable_private->xwindow,
	              gc_private->xgc, x, y+1, x + str_width, y + 1);
      /*gdk_draw_line (GTK_WIDGET (xtext)->window, col, x, y+1, x + str_width, y+1);*/

   return str_width;
}

static void
gtk_xtext_reset (GtkXText *xtext, int mark, int attribs)
{
   if (attribs)
   {
      xtext->underline = FALSE;
      xtext->bold = FALSE;
   }
   if (!mark)
   {
      xtext->backcolor = FALSE;
      XSetForeground (((GdkGCPrivate*)xtext->fgc)->xdisplay,
                      ((GdkGCPrivate*)xtext->fgc)->xgc,
                      xtext->palette[0]);
      XSetBackground (((GdkGCPrivate*)xtext->fgc)->xdisplay,
                      ((GdkGCPrivate*)xtext->fgc)->xgc,
                      xtext->palette[1]);
   }
   xtext->col_fore = 0;
   xtext->col_back = 1;
}

/* render a single line, which WONT wrap, and parse mIRC colors */

static void
gtk_xtext_render_str (GtkXText *xtext, int y, textentry *ent, char *str, int len, int win_width, int indent)
{
   GdkGC *col;
   int i = 0, x = indent, j = 0;
   char *pstr = str;
   int col_num, tmp;
   int offset;
   int mark = FALSE;
#define num xtext->num
#define nc xtext->nc

   offset = str - ent->str;

   col = xtext->fgc;

   while (i < len)
   {

      if (ent->mark_start == (i + offset))
      {
         XSetBackground (((GdkGCPrivate*)col)->xdisplay,
                         ((GdkGCPrivate*)col)->xgc,
                         xtext->palette[16]);
         XSetForeground (((GdkGCPrivate*)col)->xdisplay,
                         ((GdkGCPrivate*)col)->xgc,
                         xtext->palette[17]);
         xtext->backcolor = TRUE;
         mark = TRUE;
      }

      if ( (xtext->parsing_color && isdigit (str[i]) && nc < 2) ||
           (xtext->parsing_color && str[i] == ',' && nc < 3)        )
      {
         pstr++;
         if (str[i] == ',')
         {
            xtext->parsing_backcolor = TRUE;
            if (nc)
            {
               num[nc] = 0;
               nc = 0;
               col_num = atoi (num) % 16;
               xtext->col_fore = col_num;
               if (!mark)
                  XSetForeground (((GdkGCPrivate*)col)->xdisplay,
                              ((GdkGCPrivate*)col)->xgc,
                              xtext->palette[col_num]);
            }
         } else
         {
            num[nc] = str[i];
            if (nc < 14)
               nc++;
         }
      } else
      {
         if (xtext->parsing_color)
         {
            xtext->parsing_color = FALSE;
            if (nc)
            {
               num[nc] = 0;
               nc = 0;
               col_num = atoi (num) % 16;
               if (xtext->parsing_backcolor)
               {
                  if (col_num == 1)
                     xtext->backcolor = FALSE;
                  else
                     xtext->backcolor = TRUE;
                  if (!mark)
                     XSetBackground (((GdkGCPrivate*)col)->xdisplay,
                                 ((GdkGCPrivate*)col)->xgc,
                                 xtext->palette[col_num]);
                  xtext->col_back = col_num;
               } else
               {
                  if (!mark)
                     XSetForeground (((GdkGCPrivate*)col)->xdisplay,
                                 ((GdkGCPrivate*)col)->xgc,
                                 xtext->palette[col_num]);
                  xtext->col_fore = col_num;
               }
               xtext->parsing_backcolor = FALSE;
            } else
            {
               /* got a \003<non-digit>... i.e. reset colors */
               x += gtk_xtext_render_flush (xtext, x, y, pstr, j, col);
               pstr += j;
               j = 0;
               gtk_xtext_reset (xtext, mark, FALSE);
            }
         }

         switch (str[i])
         {
         case '\n':
            break;
         case ATTR_BEEP:
            gdk_beep ();
            break;
         case ATTR_REVERSE:
            x += gtk_xtext_render_flush (xtext, x, y, pstr, j, col);
            pstr += j+1;
            j = 0;
            tmp = xtext->col_fore;
            xtext->col_fore = xtext->col_back;
            xtext->col_back = tmp;
            if (!mark)
            {
               XSetForeground (((GdkGCPrivate*)col)->xdisplay,
                            ((GdkGCPrivate*)col)->xgc,
                            xtext->palette[xtext->col_fore]);
               XSetBackground (((GdkGCPrivate*)col)->xdisplay,
                            ((GdkGCPrivate*)col)->xgc,
                            xtext->palette[xtext->col_back]);
            }
            if (xtext->col_back != 1)
               xtext->backcolor = TRUE;
            else
               xtext->backcolor = FALSE;
            break;
         case ATTR_BOLD:
            x += gtk_xtext_render_flush (xtext, x, y, pstr, j, col);
            xtext->bold = !xtext->bold;
            pstr += j+1;
            j = 0;
            break;
         case ATTR_UNDERLINE:
            x += gtk_xtext_render_flush (xtext, x, y, pstr, j, col);
            xtext->underline = !xtext->underline;
            pstr += j+1;
            j = 0;
            break;
         case ATTR_RESET:
            x += gtk_xtext_render_flush (xtext, x, y, pstr, j, col);
            pstr += j+1;
            j = 0;
            gtk_xtext_reset (xtext, mark, TRUE);
            break;
         case ATTR_COLOR:
            x += gtk_xtext_render_flush (xtext, x, y, pstr, j, col);
            xtext->parsing_color = TRUE;
            pstr += j+1;
            j = 0;
            break;
         default:
            j++;
         }
      }
      i++;

      if (mark && ent->mark_end == (i + offset))
      {
         x += gtk_xtext_render_flush (xtext, x, y, pstr, j, col);
         pstr += j;
         j = 0;
         XSetBackground (((GdkGCPrivate*)col)->xdisplay,
                      ((GdkGCPrivate*)col)->xgc,
                      xtext->palette[xtext->col_back]);
         XSetForeground (((GdkGCPrivate*)col)->xdisplay,
                      ((GdkGCPrivate*)col)->xgc,
                      xtext->palette[xtext->col_fore]);
         if (xtext->col_back != 1)
            xtext->backcolor = TRUE;
         else
            xtext->backcolor = FALSE;
         mark = FALSE;
      }
   }

   if (j)
      x += gtk_xtext_render_flush (xtext, x, y, pstr, j, col);

   /* render the background to the right of the text */
   /*gdk_draw_rectangle (GTK_WIDGET (xtext)->window,
                       xtext->bgc,
                       1,
                       x,
                       y - (xtext->font->ascent),
                       win_width - x,
                       xtext->fontsize);*/
   XFillRectangle (((GdkWindowPrivate *)GTK_WIDGET(xtext)->window)->xdisplay,
                   ((GdkWindowPrivate *)GTK_WIDGET(xtext)->window)->xwindow,
                   ((GdkGCPrivate *)xtext->bgc)->xgc,
                   x, y - (xtext->font->ascent),
                   (win_width+1) - x, xtext->fontsize);

   if (indent)
   {     /* render background to the left of the text */
      XFillRectangle (((GdkWindowPrivate *)GTK_WIDGET(xtext)->window)->xdisplay,
                      ((GdkWindowPrivate *)GTK_WIDGET(xtext)->window)->xwindow,
                      ((GdkGCPrivate *)xtext->bgc)->xgc,
                      0, y - (xtext->font->ascent),
                      indent, xtext->fontsize);
   }
}

/* get the desktop/root window - thanks Eterm */

static Window desktop_window = None;

static Window
get_desktop_window (Window the_window)
{
  Atom prop, type, prop2;
  int format;
  unsigned long length, after;
  unsigned char *data;
  unsigned int nchildren;
  Window w, root, *children, parent;
  
  prop = XInternAtom(GDK_DISPLAY(), "_XROOTPMAP_ID", True);
  prop2 = XInternAtom(GDK_DISPLAY(), "_XROOTCOLOR_PIXEL", True);
  
  if (prop == None && prop2 == None)
    return None;
  
#ifdef WATCH_DESKTOP_OPTION
  if (Options & Opt_watchDesktop)  {
    if (TermWin.wm_parent != None)
      XSelectInput(GDK_DISPLAY(), TermWin.wm_parent, None);
    
    if (TermWin.wm_grandparent != None)
      XSelectInput(GDK_DISPLAY(), TermWin.wm_grandparent, None);
  }
#endif
  
  for (w = the_window; w; w = parent) {
    if ((XQueryTree(GDK_DISPLAY(), w, &root, &parent, &children, &nchildren)) == False) 
      return None;
      
    if (nchildren) 
      XFree(children);

#ifdef WATCH_DESKTOP_OPTION
    if (Options & Opt_watchDesktop) {
      if (w == TermWin.parent) {
	TermWin.wm_parent = parent;
	XSelectInput(GDK_DISPLAY(), TermWin.wm_parent, StructureNotifyMask);
      } else if (w == TermWin.wm_parent) {
	TermWin.wm_grandparent = parent;
	XSelectInput(GDK_DISPLAY(), TermWin.wm_grandparent, StructureNotifyMask);
      }
    }
#endif
    
    if (prop != None) {
      XGetWindowProperty(GDK_DISPLAY(), w, prop, 0L, 1L, False, AnyPropertyType,
			 &type, &format, &length, &after, &data);
    } else if (prop2 != None) {
      XGetWindowProperty(GDK_DISPLAY(), w, prop2, 0L, 1L, False, AnyPropertyType,
			 &type, &format, &length, &after, &data);
    } else  {
      continue;
    }
    
    if (type != None) {
      return (desktop_window = w);
    }
  }
  
  return (desktop_window = None);
}

/* stolen from zvt, which was stolen from Eterm */

static Pixmap
get_pixmap_prop (Window the_window, char *prop_id)
{
  Atom prop, type;
  int format;
  unsigned long length, after;
  unsigned char *data;
  
  /*this should be changed when desktop changes I guess*/
  if(desktop_window == None)
    desktop_window = get_desktop_window(the_window);
  if(desktop_window == None)
    desktop_window = GDK_ROOT_WINDOW();
  
  prop = XInternAtom(GDK_DISPLAY(), prop_id, True);
  
  if (prop == None)
    return None;
  
  XGetWindowProperty(GDK_DISPLAY(), desktop_window, prop, 0L, 1L, False,
		     AnyPropertyType, &type, &format, &length, &after,
		     &data);

  if (type == XA_PIXMAP)
    return *((Pixmap *)data);

  return None;
}

#ifdef USE_IMLIB

static GdkPixmap *
create_shaded_pixmap (Pixmap p, int x, int y, int w, int h)
{
  GdkPixmap *pix,*pp,*tmp;
  GdkImlibColorModifier mod;
  GdkImlibImage *iim;
  GdkGC *tgc;
  int width,height,depth;
  
  if (p == None)
    return NULL;

  pp = gdk_pixmap_foreign_new(p);
  gdk_window_get_geometry(pp,NULL,NULL,&width,&height,&depth);

  if (width<x+w || height<y+h || x<0 || y<0) {
    tgc = gdk_gc_new(pp);
    tmp = gdk_pixmap_new(pp,w,h,depth);
    
    gdk_gc_set_tile (tgc, pp);
    gdk_gc_set_fill (tgc, GDK_TILED);
    gdk_gc_set_ts_origin(tgc, -x, -y);
    gdk_draw_rectangle (tmp, tgc, TRUE, 0, 0, w, h);
    gdk_gc_destroy(tgc);
    
    iim = gdk_imlib_create_image_from_drawable(tmp,
					       NULL, 0, 0, w, h);
    gdk_pixmap_unref(tmp);
  } else {
    iim = gdk_imlib_create_image_from_drawable(pp,
					       NULL, x, y, w, h);
  }
  gdk_xid_table_remove (GDK_WINDOW_XWINDOW(pp));
  g_dataset_destroy (pp);
  g_free (pp);

  if(!iim)
    return NULL;

  mod.contrast = 256;
  mod.brightness = 190;
  mod.gamma = 256;
  gdk_imlib_set_image_modifier (iim, &mod);
  gdk_imlib_render (iim, iim->rgb_width, iim->rgb_height);
  pix = gdk_imlib_move_image (iim);
  gdk_imlib_destroy_image (iim);
  
  return pix;
}

#endif

/* free transparency xtext->pixmap */

static void
gtk_xtext_free_trans (GtkXText *xtext)
{
   if (xtext->pixmap)
   {
//      gdk_window_set_back_pixmap (GTK_WIDGET (xtext)->window, 0, 0);
      gdk_xid_table_remove (GDK_WINDOW_XWINDOW (xtext->pixmap));
      g_dataset_destroy (xtext->pixmap);
      g_free (xtext->pixmap);
      xtext->pixmap = NULL;
   }
}

/* grab pixmap from root window and set xtext->pixmap */

static void
gtk_xtext_load_trans (GtkXText *xtext)
{
   Pixmap rootpix;
   Window childret;
   GtkWidget *widget = GTK_WIDGET (xtext);
   int x, y;

   rootpix = get_pixmap_prop (GDK_WINDOW_XWINDOW(widget->window), "_XROOTPMAP_ID");
   if (rootpix == None)
   {
      fprintf(stderr, "XTEXT **: Cannot get root window pixmap!\n");
      xtext->transparent = FALSE;
      return;
   }

   XTranslateCoordinates (
      GDK_WINDOW_XDISPLAY (widget->window),
      GDK_WINDOW_XWINDOW (widget->window),
      GDK_ROOT_WINDOW (),
      0, 0,
      &x, &y,
      &childret);

#ifdef USE_IMLIB
   if (xtext->shaded)
   {
      int width, height;
      gdk_window_get_size(GTK_WIDGET (xtext)->window, &width, &height);
      xtext->pixmap = create_shaded_pixmap (rootpix, x, y, width, height);
      gdk_gc_set_tile (xtext->bgc, xtext->pixmap);
      gdk_gc_set_ts_origin (xtext->bgc, 0, 0);
   } else
   {
#endif
      xtext->pixmap = gdk_pixmap_foreign_new (rootpix);
      gdk_gc_set_tile (xtext->bgc, xtext->pixmap);
      gdk_gc_set_ts_origin (xtext->bgc, -x, -y);
#ifdef USE_IMLIB
   }
#endif
   gdk_gc_set_fill (xtext->bgc, GDK_TILED);
   gdk_window_set_back_pixmap (GTK_WIDGET (xtext)->window, xtext->pixmap, 0);
}

/* render a single line, which may wrap to more lines */

static int
gtk_xtext_render_line (GtkXText *xtext, textentry *ent, char *str, int line, int lines_max, int subline, int indent, int str_width)
{
   int y;
   int width;
   int height;
   int orig_len;
   int len;
   int ret = 1;

   len = orig_len = strlen (str);

   gdk_window_get_size (GTK_WIDGET (xtext)->window, &width, &height);
   width--;

   y = (xtext->fontsize * line) + xtext->font->ascent;

   if (str_width == -1)
      str_width = gtk_xtext_text_width (xtext, str, len);

   str_width += indent;

   while (str_width > width)
   {
      len--;
      if (len == 0)
         return 1;
      str_width = gtk_xtext_text_width (xtext, str, len) + indent;
   }

   if (!subline)
      gtk_xtext_render_str (xtext, y, ent, str, len, width, indent);
   else {
      subline--;
      line--;
      ret--;
   }

   if (len != orig_len && lines_max > line+1)
   {
      ret += gtk_xtext_render_line (xtext, ent, str + len, line+1, lines_max, subline, xtext->indent, -1);
   }

   return ret;
}

void
gtk_xtext_set_palette (GtkXText *xtext, GdkColor palette[])
{
   int i;

   for (i=0; i<18; i++)
      xtext->palette[i] = palette[i].pixel;

   if (GTK_WIDGET_REALIZED (xtext))
   {
      XSetForeground (((GdkGCPrivate*)xtext->fgc)->xdisplay,
                   ((GdkGCPrivate*)xtext->fgc)->xgc,
                   palette[0].pixel);
      XSetBackground (((GdkGCPrivate*)xtext->fgc)->xdisplay,
                   ((GdkGCPrivate*)xtext->fgc)->xgc,
                   palette[1].pixel);
      XSetForeground (((GdkGCPrivate*)xtext->bgc)->xdisplay,
                   ((GdkGCPrivate*)xtext->bgc)->xgc,
                   xtext->palette[1]);
      XSetWindowBackground (((GdkWindowPrivate*)GTK_WIDGET(xtext)->window)->xdisplay,
                         ((GdkWindowPrivate*)GTK_WIDGET(xtext)->window)->xwindow,
                         xtext->palette[1]);
   }
   xtext->col_fore = 0;
   xtext->col_back = 1;
}

void
gtk_xtext_set_font (GtkXText *xtext, GdkFont *font, char *name)
{
   textentry *ent;
   unsigned char i;

   if (xtext->font)
      gdk_font_unref (xtext->font);

   if (font)
   {
      xtext->font = font;
      gdk_font_ref (font);
   } else
      font = xtext->font = gdk_font_load (name);

   if (!font)
      font = xtext->font = gdk_font_load ("fixed");

   switch (font->type)
   {
      case GDK_FONT_FONT:
      {
         XFontStruct *xfont;
         xfont = (XFontStruct *)GDK_FONT_XFONT(font);
         xtext->fontsize = xtext->font->ascent + xtext->font->descent;
         if ((xfont->min_byte1 == 0) && (xfont->max_byte1 == 0))
         {
            xtext->fonttype = FONT_1BYTE;
            for (i=0; i<255; i++)
            {
               xtext->fontwidth[i] = gdk_char_width (font, i);
            }
            xtext->space_width = xtext->fontwidth[' '];
         } else
         {
            xtext->fonttype = FONT_2BYTE;
            xtext->space_width = gdk_text_width_wc (font, (GdkWChar*)"  ", 1);
         }
         break;
      }
      case GDK_FONT_FONTSET:
      {
         XFontSet fontset = (XFontSet) ((GdkFontPrivate *)font)->xfont;
         XFontSetExtents *extents = XExtentsOfFontSet(fontset);
         xtext->fontsize = extents->max_logical_extent.height;
         xtext->fonttype = FONT_SET;
         xtext->space_width = gdk_text_width_wc (font, (GdkWChar*)"  ", 1);
         break;
      }
   }


   if (xtext->indent)   /* make indent a multiple of the space width */
   {
      int i = 0;
      while (i < xtext->indent)
      {
         i += xtext->space_width;
      }
      xtext->indent = i;
   }

   xtext->xfont = (XFontStruct *)((GdkFontPrivate*)xtext->font)->xfont;

   if (GTK_WIDGET_REALIZED (xtext))
   {
      XSetFont (((GdkWindowPrivate *)GTK_WIDGET (xtext)->window)->xdisplay,
             ((GdkGCPrivate *)xtext->fgc)->xgc,
             xtext->xfont->fid);

      /* since we have a new font, we have to recalc the text widths */
      ent = xtext->text_first;
      while (ent)
      {
         ent->str_width = gtk_xtext_text_width (xtext, ent->str, ent->str_len);
         ent = ent->next;
      }

      gtk_xtext_calc_lines (xtext);
   }
}

void
gtk_xtext_set_background (GtkXText *xtext, GdkPixmap *pixmap, int shaded)
{

   if (pixmap == (void *)-1 && xtext->transparent)
   {
      xtext->shaded = shaded;
      return;
   }

   if (xtext->pixmap && !xtext->transparent)
   {
      gdk_window_set_back_pixmap (GTK_WIDGET (xtext)->window, 0, 0);
      gdk_pixmap_unref (xtext->pixmap);
      xtext->pixmap = 0;
   }

   if (xtext->pixmap && xtext->transparent)
   {
      gtk_xtext_free_trans (xtext);
   }

   if (pixmap == (void *)-1)
   {
      xtext->transparent = TRUE;
      xtext->shaded = shaded;
      if (GTK_WIDGET_REALIZED (xtext))
         gtk_xtext_load_trans (xtext);
      return;
   }

   xtext->transparent = FALSE;
   xtext->pixmap = pixmap;

   if (pixmap != 0)
   {
      gdk_pixmap_ref (pixmap);
      gdk_gc_set_tile (xtext->bgc, pixmap);
      gdk_gc_set_ts_origin (xtext->bgc, 0, 0);
      gdk_gc_set_fill (xtext->bgc, GDK_TILED);
      if (GTK_WIDGET_REALIZED (xtext))
         gdk_window_set_back_pixmap (GTK_WIDGET (xtext)->window, pixmap, 0);
   } else
   {
      if (GTK_WIDGET_REALIZED (xtext))
      {
         gdk_gc_destroy (xtext->bgc);
         xtext->bgc = gdk_gc_new (GTK_WIDGET (xtext)->window);
         XSetForeground (((GdkGCPrivate*)xtext->bgc)->xdisplay,
                   ((GdkGCPrivate*)xtext->bgc)->xgc,
                   xtext->palette[1]);
         gtk_xtext_refresh (xtext);
      }
   }
}

/* Calculate number of actual lines (with wraps), to set adj->lower. *
 * This should only be called when the window resizes.               */

static void
gtk_xtext_calc_lines (GtkXText *xtext)
{
   textentry *ent;
   int width;
   int height;
   int lines;
   int str_width;

   gdk_window_get_size (GTK_WIDGET (xtext)->window, &width, &height);
   width--;

   if (width < 30 || height < xtext->fontsize)
      return;

   lines = 0;
   ent = xtext->text_first;
   while (ent)
   {
      lines++;

      str_width = ent->str_width - width;
      while (str_width > 0)
      {
         str_width -= (width - xtext->indent);
         lines++;
      }

      ent = ent->next;
   }

   xtext->num_lines = lines;
   gtk_xtext_adjustment_set (xtext);
}

/* find the n-th line in the linked list, this includes wrap calculations */

static textentry *
gtk_xtext_nth (GtkXText *xtext, textentry *ent, int line, int width, int *subline)
{
   int lines = -1;
   int sub;
   int str_width;

   if (xtext->last_line == line) /* for quicker exposure */
   {
      *subline = xtext->last_subline;
      return xtext->last_ent;
   }

   while (ent)
   {
      lines++;
      if (line == lines)
      {
         *subline = 0;
         return ent;
      }

      sub = 0;
      str_width = ent->str_width - width;
      while (str_width > 0)
      {
         str_width -= (width - xtext->indent);
         lines++;
         sub++;
         if (line == lines)
         {
            *subline = sub;
            return ent;
         }
      }

      ent = ent->next;
   }
   return 0;
}

/* render a whole page/window, starting from 'startline' */

static void
gtk_xtext_render_page (GtkXText *xtext, int startline)
{
   textentry *ent;
   int line;
   int lines_max;
   int width;
   int height;
   int y;
   int subline;

   gdk_flush ();

   gtk_xtext_reset (xtext, FALSE, TRUE);

   gdk_window_get_size (GTK_WIDGET (xtext)->window, &width, &height);
   width--;

   if (width < 32 || height < xtext->fontsize || width < xtext->indent)
      return;

   lines_max = (height - xtext->font->descent) / xtext->fontsize;

   subline = line = 0;
   ent = xtext->text_first;

   if (startline > 0)
      ent = gtk_xtext_nth (xtext, ent, startline, width, &subline);

   xtext->last_line = startline;
   xtext->last_ent = ent;
   xtext->last_subline = subline;

   while (ent)
   {
      line += gtk_xtext_render_line (xtext, ent, ent->str, line, lines_max, subline, 0, ent->str_width);
      subline = 0;

      if (line >= lines_max)
         break;

      ent = ent->next;
   }

   /* fill the bottom of the window where there is no text                 *
    * (to overwrite any old text that was there, or to fill exposed area). *
    * This is only necessary for transparent windows, since others have    *
    * gdk_window_backing_pixmap set.                                       */
   if (xtext->transparent)
   {
      y = (xtext->fontsize * line);
      XFillRectangle (((GdkWindowPrivate *)GTK_WIDGET(xtext)->window)->xdisplay,
                      ((GdkWindowPrivate *)GTK_WIDGET(xtext)->window)->xwindow,
                      ((GdkGCPrivate *)xtext->bgc)->xgc,
                      0, y,
                      width, height - y);
      /*gdk_draw_rectangle (GTK_WIDGET (xtext)->window,
                       xtext->bgc,
                       1,
                       0,
                       y,
                       width,
                       height - y);*/
   }

   if (xtext->separator && xtext->indent)
   {
      gtk_draw_vline (GTK_WIDGET (xtext)->style,
                      GTK_WIDGET (xtext)->window, 
                      GTK_STATE_NORMAL,
                      2, GTK_WIDGET (xtext)->allocation.height-2,
                      xtext->indent - ((xtext->space_width + 1) / 2));
   }
}

static int
gtk_xtext_render_page_timeout (GtkXText *xtext)
{
   GtkAdjustment *adj = xtext->adj;
   int scroll = FALSE;

   /* is scrollbar at the bottom? */
   if (adj->value >= adj->upper - adj->page_size)
      scroll = TRUE;

   gtk_xtext_adjustment_set (xtext);

   if (scroll)
      gtk_adjustment_set_value (adj, adj->upper - adj->page_size);

   /* FIXME: page is rendered even if the new entries arn't visible */
   gtk_xtext_render_page (xtext, adj->value);

   xtext->io_tag = -1;

   return 0;
}

void
gtk_xtext_refresh (GtkXText *xtext)
{
   int win_height, win_width;

   gdk_window_get_size (GTK_WIDGET (xtext)->window, &win_width, &win_height);
   XFillRectangle (((GdkWindowPrivate *)GTK_WIDGET(xtext)->window)->xdisplay,
                   ((GdkWindowPrivate *)GTK_WIDGET(xtext)->window)->xwindow,
                   ((GdkGCPrivate *)xtext->bgc)->xgc,
                   0, 0,
                   win_width, win_height);
   gtk_xtext_render_page (xtext, xtext->adj->value);
}

void
gtk_xtext_remove_lines (GtkXText *xtext, int lines, int refresh)
{
   textentry *next;

   while (xtext->text_first && lines)
   {
      next = xtext->text_first->next;
      free (xtext->text_first);
      xtext->text_first = next;
      lines--;
   }
   if (!xtext->text_first)
      xtext->text_last = NULL;

   if (refresh)
   {
      gtk_xtext_calc_lines (xtext);
      gtk_xtext_refresh (xtext);
   }
}

void
gtk_xtext_append_indent (GtkXText *xtext,
                         char *left_text, int left_len,
                         char *right_text, int right_len)
{
   char *str;
   int spaces;
   int space;
   int i;

   if (left_len == -1)
      left_len = strlen (left_text);

   if (right_len == -1)
      right_len = strlen (right_text);

   i = 0;
   str = malloc (left_len + right_len + 128);
   str[0] = 0;

   space = xtext->indent - gtk_xtext_text_width (xtext, left_text, left_len);
   if (space > 0 && right_len)
   {
      spaces = (space / xtext->space_width) - 1;
      for (i=0; i<spaces; i++)
         str[i] = ' ';
   }

   memcpy (&str[i], left_text, left_len);
   str[i+left_len] = ' ';
   memcpy (&str[i+left_len+1], right_text, right_len);
   str[i+left_len+1+right_len] = 0;

   gtk_xtext_append (xtext, str, -1);

   free (str);   
}

void
gtk_xtext_append (GtkXText *xtext, char *text, int len)
{
   textentry *ent;
   int win_width;
   int str_width;

   if (len == -1)
      len = strlen (text);

   ent = malloc (len + 1 + sizeof (textentry));
   ent->str = (char *)ent + sizeof (textentry);
   ent->str_len = len;
   ent->str_width = gtk_xtext_text_width (xtext, text, len);
   ent->mark_start = -1;
   ent->mark_end = -1;
   memcpy (ent->str, text, len);
   ent->str[len] = 0;
   ent->next = NULL;

   if (xtext->text_last)
   {
      xtext->text_last->next = ent;
      xtext->text_last = ent;
   } else
   {
      xtext->text_first = ent;
      xtext->text_last = ent;
   }

   win_width = ((GdkWindowPrivate*)(GTK_WIDGET (xtext)->window))->width;
   xtext->num_lines++;
   str_width = ent->str_width - win_width;
   while (str_width > 0)
   {
      str_width -= (win_width - xtext->indent);
      xtext->num_lines++;
   }

   if (xtext->max_lines > 0 && xtext->max_lines < xtext->num_lines)
   {
      gtk_xtext_remove_lines (xtext, 1, FALSE);
   }

   if (xtext->frozen == 0 && xtext->io_tag == -1)
   {
      xtext->io_tag = gtk_timeout_add (REFRESH_TIMEOUT,
                                       (GtkFunction)gtk_xtext_render_page_timeout,
                                       xtext);
   }
}

gchar* gtk_xtext_get_chars(GtkXText *xtext)
{
    int lenght=0;
    gchar *chars;
    textentry *tentry=xtext->text_first;
    while(tentry!=NULL)
    {
        lenght+=tentry->str_len;
        tentry=tentry->next;
    }
    if (lenght==0) return NULL;
    chars=g_malloc0(lenght);

    tentry=xtext->text_first;
    while(tentry!=NULL)
    {
        strcat(chars,tentry->str);
        tentry=tentry->next;
    }

    return chars;
}
